// -*- C++ -*-

//=============================================================================
/**
 *  @file    analyzer.cpp
 *
 *  analyzer.cpp,v 1.2 2003/08/25 07:01:08 jwillemsen Exp
 *
 *  This file contains a DynAny analyzer that dumps the contents of a DynAny
 *  to the logging.
 *
 *  @author  Johnny Willemsen  (jwillemsen@remedy.nl)
 */
//=============================================================================

#include "analyzer.h"

#define CASEE(type,CT,str) case CORBA::tk_##type: {\
  CORBA::CT b = da->get_##type() ; \
  tab(level_);\
  if (debug_) \
    ACE_DEBUG ((LM_DEBUG, str , b)); \
  } break;

DynAnyAnalyzer::DynAnyAnalyzer(CORBA::ORB_ptr orb, DynamicAny::DynAnyFactory_ptr dynany_factory, int debug)
  : orb_ (CORBA::ORB::_duplicate(orb)),
    dynany_factory_ (DynamicAny::DynAnyFactory::_duplicate(dynany_factory)),
    level_ (0),
    debug_ (debug),
    INDENT (4) 

{
}

void DynAnyAnalyzer::tab(int t)
{
  if (debug_)
    {
      for( int i = 0 ; i < t ; i++ )
          ACE_DEBUG ((LM_DEBUG,
                      "\t"));
    }
}

DynAnyAnalyzer::~DynAnyAnalyzer()
{
}

void DynAnyAnalyzer::resetTab()
{
  level_ = 0;
}

void DynAnyAnalyzer::analyze (DynamicAny::DynAny_ptr da ACE_ENV_ARG_DECL)
{
  CORBA::TypeCode_var tc = da->type();

  ACE_DEBUG ((LM_DEBUG,
               "\n\n Show_TC begin ********\n\n\n"));


  this->show_TC(tc.in());

  ACE_DEBUG ((LM_DEBUG,
               "\n\n Show_TC end   ********\n\n\n"));


  // strip aliases
  while( tc->kind() == CORBA::tk_alias )
    {
      tc = tc->content_type();
    }

  switch( tc->kind() )
   {
     case CORBA::tk_struct :
      {
        DynamicAny::DynStruct_var ds
          = DynamicAny::DynStruct::_narrow(da ACE_ENV_ARG_PARAMETER);
        ACE_CHECK;

        tab(level_);

        if (debug_)
          ACE_DEBUG ((LM_DEBUG,
                     "STRUCT\n"));

        if( da->seek(0) )
          {
            level_++;
            do
              {
                DynamicAny::DynAny_var cc    = ds->current_component(ACE_ENV_SINGLE_ARG_PARAMETER);
                ACE_CHECK;

                DynamicAny::FieldName_var fn = ds->current_member_name(ACE_ENV_SINGLE_ARG_PARAMETER);
                ACE_CHECK;

                tab(level_);

                if (debug_)
                  ACE_DEBUG ((LM_DEBUG,
                             "Member = %s\n", fn.in()));

                if (!CORBA::is_nil (cc.in ()))
                  {
                    this->analyze (cc.in() ACE_ENV_ARG_PARAMETER);
                    ACE_CHECK;
                  }

              } while( da->next() );
            level_--;
          }
      }
      break; // end tk_struct

     case CORBA::tk_sequence:
       {
         DynamicAny::DynSequence_var ds
           = DynamicAny::DynSequence::_narrow(da ACE_ENV_ARG_PARAMETER);
        ACE_CHECK;

        int i = 0;

        tab(level_);

        if (debug_)
          ACE_DEBUG ((LM_DEBUG,
                      "SEQUENCE\n"));

        if( ds->seek(0) )
          {
            level_++;
            do
              {
                tab(level_);

                if (debug_)
                  ACE_DEBUG ((LM_DEBUG,
                              "[%d]\n", i));

                DynamicAny::DynAny_var cc( ds->current_component() );
                ACE_CHECK;

                if (!CORBA::is_nil (cc.in ()))
                  {
                    analyze(cc.in() ACE_ENV_ARG_PARAMETER);
                    ACE_CHECK;
                  }

                i++;
              } while( da->next() );
            level_--;
         }
       }
       break; // end tk_sequence

     case CORBA::tk_array:
       {
          tab(level_);

          if (debug_)
            ACE_DEBUG ((LM_DEBUG,
                       "ARRAY\n"));

          level_++;
          for( unsigned int i = 0 ; i < tc->length() ; i++ )
            {
              tab(level_);

              if (debug_)
                ACE_DEBUG ((LM_DEBUG,
                            "[%d]\n", i));

              DynamicAny::DynAny_var cc = da->current_component();

              if (!CORBA::is_nil (cc.in ()))
                {
                  analyze(cc.in() ACE_ENV_ARG_PARAMETER);
                  ACE_CHECK;
                }

              da->next();
            }
          level_--;
       }
       break;

     case CORBA::tk_union:
       {
         DynamicAny::DynUnion_var value
           = DynamicAny::DynUnion::_narrow(da ACE_ENV_ARG_PARAMETER);
         ACE_CHECK;

         if( !value->has_no_active_member() )
           {
             DynamicAny::DynAny_var disc = value->member();

              if (!CORBA::is_nil (disc.in ()))
                {
                  this->analyze(disc.in() ACE_ENV_ARG_PARAMETER);
                  ACE_CHECK;
                }
           }
       }
       break;

     case CORBA::tk_any:
       {
         DynamicAny::DynAny_var dynany;
         CORBA::Any_var any = da->get_any();

         dynany = dynany_factory_->create_dyn_any(any.in() ACE_ENV_ARG_PARAMETER);
         ACE_CHECK;

         if (!CORBA::is_nil (dynany.in ()))
           {
             this->analyze(dynany.in() ACE_ENV_ARG_PARAMETER);
             ACE_CHECK;
           }

         dynany->destroy();
       }
       break;

     case CORBA::tk_enum:
       {
         DynamicAny::DynEnum_var value
           = DynamicAny::DynEnum::_narrow(da ACE_ENV_ARG_PARAMETER);
         ACE_CHECK;

          CORBA::String_var s = value->get_as_string();
          tab(level_);

          if (debug_)
            ACE_DEBUG ((LM_DEBUG,
                       "  Value (enum) = %s\n", s.in()));
       }
       break;

     CASEE(boolean,Boolean,"  Value (bool) = %d\n");
     CASEE(short,Short,"  Value (short) = %d\n");
     CASEE(ushort,UShort,"  Value (ushort) = %u\n");
     CASEE(long,Long,"  Value (long) = %d\n");
     CASEE(ulong,ULong,"  Value (ulong) = %u\n");
     CASEE(longlong,LongLong,"  Value (longlong) %Ld\n");
     CASEE(ulonglong,ULongLong,"  Value (ulonglong) %Lu\n");
     CASEE(char,Char,"  Value (char) = %c\n");
     CASEE(float,Float,"  Value (float) = %f\n");
     CASEE(double,Double,"  Value (double) = %f\n");
     CASEE(octet,Octet,"  Value (octet) = %c\n");

     case CORBA::tk_string:
       {
         CORBA::String_var b( da->get_string() );

         tab(level_);
         if (debug_)
           ACE_DEBUG ((LM_DEBUG,
                      "  Value (string) = %s\n", b.in()));
       }
       break;

     case CORBA::tk_TypeCode:
       {
         tab(level_);
         if (debug_) \
           ACE_DEBUG ((LM_DEBUG,
                       "  Value (TypeCode) = %d\n" , da->get_typecode()->kind()));
       }
       break;

     default:
       {
         tab(level_);
         if (debug_)
           ACE_DEBUG ((LM_DEBUG,
                      "  unhandled typecode = %d\n", (int) tc->kind()));
       }
       break;
   }

   if (debug_)
     ACE_DEBUG ((LM_DEBUG,
                 "\n"));
}


// show_TC start


void DynAnyAnalyzer::indent(int indent_lvl)
{
    for (int i = 0; i < INDENT * indent_lvl; i++)
        cout.put(' ');
}

void DynAnyAnalyzer::show_label(const CORBA::Any * ap)
{

    CORBA::TypeCode_var tc = ap->type();
    if (tc->kind() == CORBA::tk_octet) {
        cout << "default:" << endl;
    } else {
        cout << "case ";
        switch (tc->kind()) {
        case CORBA::tk_short:
            CORBA::Short s;
            *ap >>= s;
            cout << s;
            break;
        case CORBA::tk_long:
            CORBA::Long l;
            *ap >>= l;
            cout << l;
            break;
        case CORBA::tk_ushort:
            CORBA::UShort us;
            *ap >>= us;
            cout << us;
            break;
        case CORBA::tk_ulong:
            CORBA::ULong ul;
            *ap >>= ul;
            cout << ul;
            break;
        case CORBA::tk_boolean:
            CORBA::Boolean b;
            *ap >>= CORBA::Any::to_boolean(b);
            cout << (b ? "TRUE" : "FALSE");
            break;
        case CORBA::tk_char:
            CORBA::Char c;
            *ap >>= CORBA::Any::to_char(c);
            if (isalnum(c)) {
                cout << "'" << c << "'";
            } else {
                cout << "'\\" << setw(3) << setfill('0')
                     << oct << (unsigned)c << "'";
            }
            break;
        case CORBA::tk_longlong:
            CORBA::LongLong ll;
            *ap >>= ll;
            cout << ll;
            break;
        case CORBA::tk_ulonglong:
            CORBA::ULongLong ull;
            *ap >>= ull;
            cout << ull;
            break;
        case CORBA::tk_wchar:
            CORBA::WChar wc;
            *ap >>= CORBA::Any::to_wchar(wc);
            cout << "'" << wc << "'";
            break;
        case CORBA::tk_enum:
            {
                DynamicAny::DynAny_var da = dynany_factory_->create_dyn_any(*ap);
                DynamicAny::DynEnum_var de = DynamicAny::DynEnum::_narrow(da);
                CORBA::String_var s = de->get_as_string();
                cout << s;
                da->destroy();
            }
            break;
        default:
            // Union discriminator can't be anything else
            abort();
        }
        cout << ":" << endl;
    }
}


void DynAnyAnalyzer::show_TC(
    CORBA::TypeCode_ptr         tcp,
    list<CORBA::TypeCode_var> & tlist,
    int                         indent_lvl)
{
    static const char * const kind_name[] = {
        "tk_null", "void", "short", "long",
        "unsigned short", "unsigned long", "float",
        "double", "boolean", "char", "octet", "any",
        "CORBA::TypeCode", "CORBA::Principal",
        "interface", "struct", "union", "enum",
        "string", "sequence", "array", "typedef",
        "exception", "long long", "unsigned long long",
        "long double", "wchar", "wstring", "fixed"
    };

    indent(indent_lvl);
    
    cout << kind_name[tcp->kind()];     // Print the TCKind value.

    //
    // Print name and repository ID for those type codes
    // that have these parameters.
    //
    switch (tcp->kind()) {
    case CORBA::tk_objref:
    case CORBA::tk_struct:
    case CORBA::tk_union:
    case CORBA::tk_except:
    case CORBA::tk_enum:
    case CORBA::tk_alias:
        cout << " " << tcp->name()
             << " (" << tcp->id() << "):" << endl;
    default:
        ;       // Do nothing
    }

    //
    // For type codes that have other parameters,
    // show the remaining parameters.
    //
    switch (tcp->kind()) {
    default:                    // No other params to print
        cout << endl;
        break;
    //
    // For fixed types, show digits and scale.
    //
    case CORBA::tk_fixed:
        cout << "<" << tcp->fixed_digits() << ","
             << tcp->fixed_scale() << ">" << endl;
        break;
    //
    // For enumerations, show the enumerators.
    //
    case CORBA::tk_enum:
        {
            indent(indent_lvl + 1);
            for (CORBA::ULong i = 0; i < tcp->member_count(); i++) {
                cout << tcp->member_name(i);
                if (i < tcp->member_count() - 1)
                    cout << ", ";
            }
            cout << endl;
        }
        break;
    //
    // For strings, show the bound (if any).
    //
    case CORBA::tk_string:
    case CORBA::tk_wstring:
        {
            CORBA::ULong l = tcp->length();
            if (l != 0)
                cout << "<" << l << ">";
            cout << endl;
        }
        break;
    //
    // For sequences, show the bound (if any) and
    // the element type.
    //
    case CORBA::tk_sequence:
        {
            CORBA::ULong l = tcp->length();
            if (l != 0)
                cout << "<" << l << ">";
            cout << ":" << endl;
            CORBA::TypeCode_var etype = tcp->content_type();
            show_TC(etype, tlist, indent_lvl + 1);
        }
        break;
    //
    // For arrays, show the dimension and element type.
    //
    case CORBA::tk_array:
        {
            CORBA::ULong l = tcp->length();
            cout << "[" << l << "]:" << endl;
            CORBA::TypeCode_var etype = tcp->content_type();
            show_TC(etype, tlist, indent_lvl + 1);
        }
        break;
    //
    // For typedefs, show the type of the aliased type.
    //
    case CORBA::tk_alias:
        {
            CORBA::TypeCode_var atype = tcp->content_type();
            show_TC(atype, tlist, indent_lvl + 1);
        }
        break;
    //
    // For structures and exceptions, show the
    // names and types of each member.
    //
    case CORBA::tk_struct:
    case CORBA::tk_except:
        {
            //
            // Avoid a recursive loop by checking whether we
            // have shown this type code before.
            //
            list<CORBA::TypeCode_var>::iterator where;
            where = find_if(
                        tlist.begin(),
                        tlist.end(),
                        EqualTypeCodes(tcp)
                    );
            //
            // If we have not seen this type code before, add it
            // to the list of type codes processed so far and
            // decode the member type codes.
            //
            if (where == tlist.end()) {
                tlist.push_back(CORBA::TypeCode::_duplicate(tcp));
                for (CORBA::ULong i = 0; i < tcp->member_count(); i++) {
                    cout << tcp->member_name(i) << ":" << endl;
                    indent(indent_lvl + 1);
                    CORBA::TypeCode_var mt = tcp->member_type(i);
                    show_TC(mt, tlist, indent_lvl + 2);
                }
            } else {
                cout << " " << tcp->name()
                     << " (" << tcp->id() << ")" << endl;
            }
        }
        break;
    //
    // For unions, show the discriminator type.
    // Then, for each member, show the case label,
    // member name, and member type. To show the case
    // label, we use the show_label() helper function.
    //
    case CORBA::tk_union:
        {
            //
            // Avoid getting trapped in a recursive loop.
            //
            list<CORBA::TypeCode_var>::iterator where;
            where = find_if(
                        tlist.begin(),
                        tlist.end(),
                        EqualTypeCodes(tcp)
                    );
            //
            // Show the members only if we haven't shown this type
            // code before.
            //
            if (where == tlist.end()) {
                tlist.push_back(CORBA::TypeCode::_duplicate(tcp));
                indent(indent_lvl + 1);

                //
                // Show discriminator type.
                //
                cout << "Discriminator type:" << endl;
                CORBA::TypeCode_var dt;
                dt = tcp->discriminator_type();
                show_TC(dt, tlist, indent_lvl + 2);

                //
                // Show case label, member name, and
                // member type for each member.
                //
                for (CORBA::ULong i = 0; i < tcp->member_count(); i++) {
                    CORBA::Any_var label = tcp->member_label(i);
                    indent(indent_lvl + 1);
                    show_label(label);
                    indent(indent_lvl + 2);
                    cout << tcp->member_name(i) << ":" << endl;
                    CORBA::TypeCode_var mt = tcp->member_type(i);
                    show_TC(mt, tlist, indent_lvl + 3);
                }
            } else {
                cout << " " << tcp->name()
                     << " (" << tcp->id() << ")" << endl;
            }
        }
        break;
    }
}


void DynAnyAnalyzer::show_TC(CORBA::TypeCode_ptr tcp)
{
    list<CORBA::TypeCode_var> tclist;
    show_TC(tcp, tclist, 0);
}

// show_TC end
